from __future__ import annotations
from typing import List, Optional
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.dataset_versions import DatasetVersionManager
from continual.python.sdk.promotions import PromotionManager
from continual.python.sdk.tags import TagsManager
from continual.python.sdk.metadata import MetadataManager


from continual.python.sdk.batchpredictions import (
    BatchPredictionManager,
)


class DatasetManager(Manager):
    """Manages dataset resources."""

    name_pattern: str = (
        "projects/{project}/environments/{environment}/datasets/{dataset}"
    )

    def create(
        self,
        display_name: str,
        description: Optional[str] = "",
        get_if_exists: bool = True,
    ) -> Dataset:
        """Create dataset.

        Arguments:
            display_name: Dataset name or id.
            description: A brief description of this dataset.
            get_if_exists: If True, this call will get the dataset with the display name, else throw an error.

        Returns
            A Dataset.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> dataset = run.datasets.create(display_name="my_dataset", description="Customer churn dataset")
            <Dataset object {'name': 'projects/test_proj_2/environments/test_env/datasets/my_dataset',
            'description': 'Customer churn dataset', 'author': 'users/BefwyWcn6x7SNC533zfaAR',
            'display_name': 'my_dataset', 'create_time': '2022-12-19T06:06:34.763763Z', 'update_time': '2022-12-19T06:06:34.763763Z',
            'state': '', 'current_version': ''}>
        """
        req = management_pb2.CreateDatasetRequest(
            name=self.name(display_name),
            description=description,
            get_if_exists=get_if_exists,
            run=self.run_name,
        )
        resp = self.client._management.CreateDataset(req)
        return Dataset.from_proto(resp, client=self.client, parent_run=self.run_name)

    def get(self, id: str) -> Dataset:
        """Get dataset.

        Arguments:
            id: Dataset name or id.

        Returns
            An experiment.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> dataset = run.datasets.get("my_dataset")
            <Dataset object {'name': 'projects/test_proj_2/environments/test_env/datasets/my_dataset',
            'description': 'Customer churn dataset', 'author': 'users/BefwyWcn6x7SNC533zfaAR', 'display_name': 'my_dataset',
            'create_time': '2022-12-19T06:06:34.763763Z', 'update_time': '2022-12-19T06:06:34.763763Z',
            'state': '', 'current_version': ''}>
        """
        req = management_pb2.GetDatasetRequest(name=self.name(id))
        resp = self.client._management.GetDataset(req)
        return Dataset.from_proto(resp, client=self.client, parent_run=self.run_name)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: str = None,
        latest: bool = True,
        all_projects: bool = False,
    ) -> List[Dataset]:
        """List dataset.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.
            all_projects:  Whether to include all instances of this resource from the project or just from the current parent.

        Returns:
            A list of Datasets.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> ds = [run.datasets.create(display_name=f"my_dataset_{i}", description=f"Customer churn dataset {i}") for i in range(10)]
            >>> [r.display_name for r in run.datasets.list(page_size=10)]
            ['my_dataset_9', 'my_dataset_8', 'my_dataset_7', 'my_dataset_6', 'my_dataset_5', 'my_dataset_4', 'my_dataset_3', 'my_dataset_2', 'my_dataset_1', 'my_dataset_0']
            >>> [r.display_name for r in run.datasets.list(page_size=10, latest=False)]
            ['my_dataset_0', 'my_dataset_1', 'my_dataset_2', 'my_dataset_3', 'my_dataset_4', 'my_dataset_5', 'my_dataset_6', 'my_dataset_7', 'my_dataset_8', 'my_dataset_9']
            >>> [r.display_name for r in run.datasets.list(page_size=10, order_by="display_name")]
            ['my_dataset_0', 'my_dataset_1', 'my_dataset_2', 'my_dataset_3', 'my_dataset_4', 'my_dataset_5', 'my_dataset_6', 'my_dataset_7', 'my_dataset_8', 'my_dataset_9']
        """
        req = management_pb2.ListDatasetsRequest(
            parent=self.parent,
            page_size=page_size,
            all_projects=all_projects,
            order_by=order_by,
            latest=latest,
        )
        resp = self.client._management.ListDatasets(req)
        return [
            Dataset.from_proto(x, client=self.client, parent_run=self.run_name)
            for x in resp.datasets
        ]

    def list_all(self) -> Pager[Dataset]:
        """List all dataset.

        Pages through all datasets using an iterator.

        Returns:
            A iterator of all dataset.

        Examples:
            >>> ... # Assume client, project, and environment are defined.
            >>> env = project.environments.get("my_environment")
            >>> run = env.runs.create("my_run")
            >>> ds = [run.datasets.create(display_name=f"my_dataset_{i}", description=f"Customer churn dataset {i}") for i in range(10)]
            >>> [r.display_name for r in run.datasets.list_all()]
            ['my_dataset_0', 'my_dataset_1', 'my_dataset_2', 'my_dataset_3', 'my_dataset_4', 'my_dataset_5', 'my_dataset_6', 'my_dataset_7', 'my_dataset_8', 'my_dataset_9']
        """

        def next_page(next_page_token):
            req = management_pb2.ListDatasetsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListDatasets(req)
            return (
                [
                    Dataset.from_proto(x, client=self.client, parent_run=self.run_name)
                    for x in resp.datasets
                ],
                resp.next_page_token,
            )

        return Pager(next_page)


class Dataset(Resource, types.Dataset):
    """Dataset resource."""

    name_pattern: str = (
        "projects/{project}/environments/{environment}/datasets/{dataset}"
    )
    _manager: DatasetManager
    """Dataset Manager."""

    _dataset_versions: DatasetVersionManager
    """Dataset version manager."""

    _batch_predictions: BatchPredictionManager
    """Batch Prediction manager."""

    _tags: TagsManager
    """Tags Manager"""

    _metadata: MetadataManager
    """Metadata Manager"""

    def _init(self):
        self._manager = DatasetManager(
            parent=self.parent, client=self.client, run_name=self.parent_run
        )
        self._dataset_versions = DatasetVersionManager(
            parent=self.name, client=self.client, run_name=self.parent_run
        )
        self._batch_predictions = BatchPredictionManager(
            parent=self.name, client=self.client
        )
        self._tags = TagsManager(
            parent=self.name, client=self.client, run_name=self.parent_run
        )
        self._metadata = MetadataManager(
            parent=self.name, client=self.client, run_name=self.parent_run
        )

    @property
    def dataset_versions(self) -> DatasetVersionManager:
        """Dataset version manager."""
        return self._dataset_versions

    @property
    def promotions(self) -> PromotionManager:
        """Promotion manager."""
        return self._promotions

    @property
    def batch_predictions(self) -> BatchPredictionManager:
        """Batch Prediction manager."""
        return self._batch_predictions

    @property
    def tags(self) -> TagsManager:
        """Tags Manager"""
        return self._tags

    @property
    def metadata(self) -> MetadataManager:
        """Metadata Manager"""
        return self._metadata



from __future__ import annotations
from typing import List, Optional
from continual.python.sdk.artifacts import ArtifactsManager
from continual.python.sdk.datasets import DatasetManager
from continual.python.sdk.runs import RunManager
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.models import ModelManager


class EnvironmentManager(Manager):
    """Manages environment resources."""

    name_pattern: str = "projects/{project}/environments/{environment}"

    def create(
        self, id: str, source: str = "", get_if_exists: bool = True
    ) -> Environment:
        """Create an environment.

        New environments are identified by a unique ID within
        their parent project.

        Arguments:
            id: Environment ID.
            source: Metadata about how this environment was created.
            get_if_exists: If True and the environment already exists, return the existing
                environment.  If False and the environment already exists, raise an error.

        Returns:
            A new environment.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> proj = client.projects.create(display_name="example-proj", organization=org.name)
            >>> proj.environments.create(id="example")
            <Environment object {'name': 'projects/example_proj_1/environments/example',
            'summary': {'feature_set_health': {'healthy_count': 0, 'unhealthy_count': 0, 'critical_count': 0},
            'model_health': {'healthy_count': 0, 'unhealthy_count': 0, 'critical_count': 0}, 'feature_set_count': 0,
            'feature_set_row_count': '0', 'feature_set_bytes': '0', 'feature_count': 0, 'connection_count': 0,
            'model_count': 0, 'model_version_count': 0, 'experiment_count': '0', 'prediction_count': '0'},
            'data_store': {'type': '', 'is_demo_data_store': False}, 'update_time': '2022-12-15T23:47:28.641932Z',
            'create_time': '2022-12-15T23:47:28.641932Z', 'scheduling_enabled': False, 'source': '', 'env_type': '',
            'is_protected': False, 'is_ephemeral': False}>
            >>> env = proj.environments.create(id="example")) # get_if_exists is True by default
            >>> env = proj.environments.create(id="example", get_if_exists=False) # Raises an error
            continual.python.sdk.exceptions.AlreadyExistsError: ('Resource already exists.', {'name': 'Environment with ID (example) already exists.'})
        """
        req = management_pb2.CreateEnvironmentRequest(
            parent=self.parent,
            environment_id=id,
            source=source,
            get_if_exists=get_if_exists,
        )
        resp = self.client._management.CreateEnvironment(req)
        return Environment.from_proto(resp, client=self.client)

    def name(self, id: str, parent: Optional[str] = None) -> str:
        """Get the fully qualified name of this environment.

        Arguments:
            id: The environment id.
            parent: The parent project name.

        Return:
            string name of the environment.
        """
        if "/" in id:
            # Don't allow names to override manager parent config since this is confusing
            # and is typically a bug in the user code.
            if parent is not None and parent != "" and not id.startswith(parent):
                raise ValueError(f"Resource {id} not a child of {parent}.")
            return id

        name_str = self.parent or ""
        name_str += "/environments/" + id
        return name_str

    def get(self, id: str = "production") -> Environment:
        """Get environment.

        Arguments:
            id: environment name or id.

        Returns
            A environment.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> proj = client.projects.create(display_name="example-proj", organization=org.name)
            >>> env = proj.environments.create(id="example")
            >>> proj.environments.get(id=env.name)
            <Environment object {'name': 'projects/example_proj_1/environments/example',
            'summary': {'feature_set_health': {'healthy_count': 0, 'unhealthy_count': 0, 'critical_count': 0},
            'model_health': {'healthy_count': 0, 'unhealthy_count': 0, 'critical_count': 0}, 'feature_set_count': 0,
            'feature_set_row_count': '0', 'feature_set_bytes': '0', 'feature_count': 0, 'connection_count': 0,
            'model_count': 0, 'model_version_count': 0, 'experiment_count': '0', 'prediction_count': '0'},
            'data_store': {'type': '', 'is_demo_data_store': False}, 'update_time': '2022-12-15T23:47:28.641932Z',
            'create_time': '2022-12-15T23:47:28.641932Z', 'scheduling_enabled': False, 'source': '', 'env_type': '',
            'is_protected': False, 'is_ephemeral': False}>
        """
        req = management_pb2.GetEnvironmentRequest(name=self.name(id))
        resp = self.client._management.GetEnvironment(req)
        return Environment.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        order_by: Optional[str] = None,
        latest: bool = True,
    ) -> List[Environment]:
        """List environments.

        Arguments:
            page_size: Number of items to return.
            order_by: A string field name used to order list.
            latest: If true, the results are sorted in descending order, else ascending.

        Returns:
            A list of environments.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> proj = client.projects.create(display_name="example-proj", organization=org.name)
            >>> envs = [proj.environments.create(id=f"env{i}") for i in range(3)]
            >>> [env.name for env in proj.environments.list(page_size=5, latest=True)]
            ['projects/example_proj_2/environments/env2', 'projects/example_proj_2/environments/env1',
            'projects/example_proj_2/environments/env0', 'projects/example_proj_2/environments/production']
        """
        req = management_pb2.ListEnvironmentsRequest(
            parent=self.parent,
            page_size=page_size,
            order_by=order_by,
            latest=latest,
        )
        resp = self.client._management.ListEnvironments(req)
        return [
            Environment.from_proto(x, client=self.client) for x in resp.environments
        ]

    def list_all(self) -> Pager[Environment]:
        """List all environments.

        Pages through all environments using an iterator.

        Returns:
            A iterator of all environments.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> proj = client.projects.create(display_name="example-proj", organization=org.name)
            >>> envs = [proj.environments.create(id=f"env{i}") for i in range(3)]
            >>> [env.name for env in proj.environments.list_all()]
            ['projects/example_proj_2/environments/env0', 'projects/example_proj_2/environments/env1',
            'projects/example_proj_2/environments/env2', 'projects/example_proj_2/environments/production']
        """

        def next_page(next_page_token):
            req = management_pb2.ListEnvironmentsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListEnvironments(req)
            return (
                [
                    Environment.from_proto(x, client=self.client)
                    for x in resp.environments
                ],
                resp.next_page_token,
            )

        return Pager(next_page)

    def delete(
        self,
        id: str,
    ) -> None:
        """Delete an Environment.

        Arguments:
            id: Environment name or id.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> proj = client.projects.create(display_name="example-proj", organization=org.name)
            >>> env = proj.environments.create(id="example")
            >>> len(proj.environments.list_all())
            1
            >>> proj.environments.delete(id=env.name)
            >>> len(proj.environments.list_all())
            0
        """

        req = management_pb2.DeleteEnvironmentRequest(name=self.name(id))
        self.client._management.DeleteEnvironment(req)


class Environment(Resource, types.Environment):
    """Environment resource."""

    name_pattern: str = "projects/{project}/environments/{environment}"
    _manager: EnvironmentManager

    _models: ModelManager

    _datasets: DatasetManager

    _artifacts: ArtifactsManager

    _runs: RunManager

    def _init(self):
        self._manager = EnvironmentManager(parent=self.parent, client=self.client)
        self._models = ModelManager(parent=self.name, client=self.client)
        self._datasets = DatasetManager(parent=self.name, client=self.client)
        self._artifacts = ArtifactsManager(parent=self.name, client=self.client)
        self._runs = RunManager(parent=self.name, client=self.client)

    @property
    def models(self) -> ModelManager:
        """Models manager."""
        return self._models

    @property
    def datasets(self) -> DatasetManager:
        """Datasets manager."""
        return self._datasets

    @property
    def artifacts(self) -> ArtifactsManager:
        """Artifacts manager."""
        return self._artifacts

    @property
    def runs(self) -> RunManager:
        """Runs manager."""
        return self._runs

    def delete(self) -> None:
        """Delete environment.

        Examples:
            >>> from continual import Client
            >>> client = Client() # Assuming credentials in YAML config
            >>> org = client.organizations.create(display_name="example-org")
            >>> proj = client.projects.create(display_name="example-proj", organization=org.name)
            >>> env = proj.environments.create(id="example")
            >>> len(proj.environments.list_all())
            1
            >>> env.delete()
            >>> len(proj.environments.list_all())
            0
        """
        self._manager.delete(self.name)
